/*
 * Copyright 2020 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.google.auto.factory.otherpackage;

import com.google.auto.factory.AutoFactory;
import com.google.auto.factory.Provided;
import com.google.auto.factory.ReferencePackageFactory;

@AutoFactory
public class OtherPackage {
  private final ReferencePackageFactory referencePackageFactory;
  private final int random;

  public OtherPackage(@Provided ReferencePackageFactory referencePackageFactory, int random) {
    this.referencePackageFactory = referencePackageFactory;
    this.random = random;
  }

  public ReferencePackageFactory referencePackageFactory() {
    return referencePackageFactory;
  }

  public int random() {
    return random;
  }
}
